
/* : : generated by proto : : */

#if !defined(__PROTO__)
#include <prototyped.h>
#endif
#include "module.h"
#include "network.h"
#include "misc.h"

#include "servers.h"

/* Returns TRUE if mask contains '!' ie. address should be checked too.
   Also checks if mask contained any wildcards. */
static int check_address __PARAM__((const char *mask, int *wildcards), (mask, wildcards)) __OTORP__(const char *mask; int *wildcards;){
	int ret;

	ret = FALSE;
	while (*mask != '\0') {
		if (*mask == '!') {
			if (*wildcards) return TRUE;
			ret = TRUE;
		}

		if (*mask == '?' || *mask == '*') {
			*wildcards = TRUE;
			if (ret) return TRUE;
		}
		mask++;
	}

	return ret;
}

static int check_mask __PARAM__((SERVER_REC *server, const char *mask,
		      const char *str, int wildcards), (server, mask, str, wildcards)) __OTORP__(SERVER_REC *server; const char *mask;
		      const char *str; int wildcards;){
	if (server != NULL && server->mask_match_func != NULL) {
		/* use server specified mask match function */
		return server->mask_match_func(mask, str);
	}

	return wildcards ? match_wildcards(mask, str) :
		g_strcasecmp(mask, str) == 0;
}

int mask_match __PARAM__((SERVER_REC *server, const char *mask,
	       const char *nick, const char *user, const char *host), (server, mask, nick, user, host)) __OTORP__(SERVER_REC *server; const char *mask;
	       const char *nick; const char *user; const char *host;){
	char *str;
	int ret, wildcards;

	g_return_val_if_fail(server == NULL || IS_SERVER(server), FALSE);
	g_return_val_if_fail(mask != NULL && nick != NULL &&
			     nick != NULL && host != NULL, FALSE);

	str = !check_address(mask, &wildcards) ? (char *) nick :
		g_strdup_printf("%s!%s@%s", nick, user, host);
	ret = check_mask(server, mask, str, wildcards);
	if (str != nick) g_free(str);

	return ret;
}

int mask_match_address __PARAM__((SERVER_REC *server, const char *mask,
		       const char *nick, const char *address), (server, mask, nick, address)) __OTORP__(SERVER_REC *server; const char *mask;
		       const char *nick; const char *address;){
	char *str;
	int ret, wildcards;

	g_return_val_if_fail(server == NULL || IS_SERVER(server), FALSE);
	g_return_val_if_fail(mask != NULL && nick != NULL, FALSE);
	if (address == NULL) address = "";

	str = !check_address(mask, &wildcards) ? (char *) nick :
		g_strdup_printf("%s!%s", nick, address);
	ret = check_mask(server, mask, str, wildcards);
	if (str != nick) g_free(str);

	return ret;
}

int masks_match __PARAM__((SERVER_REC *server, const char *masks,
		const char *nick, const char *address), (server, masks, nick, address)) __OTORP__(SERVER_REC *server; const char *masks;
		const char *nick; const char *address;){
	int (*mask_match_func) __PROTO__((const char *, const char *));
	char **list, **tmp, *mask;
	int found;

	g_return_val_if_fail(server == NULL || IS_SERVER(server), FALSE);
	g_return_val_if_fail(masks != NULL &&
			     nick != NULL && address != NULL, FALSE);

	if (*masks == '\0')
                return FALSE;

	mask_match_func = server != NULL && server->mask_match_func != NULL ?
		server->mask_match_func : match_wildcards;

	found = FALSE;
	mask = g_strdup_printf("%s!%s", nick, address);
	list = g_strsplit(masks, " ", -1);
	for (tmp = list; *tmp != NULL; tmp++) {
		if (g_strcasecmp(*tmp, nick) == 0) {
                        found = TRUE;
			break;
		}

		if (mask_match_func(*tmp, mask)) {
			found = TRUE;
			break;
		}
	}
	g_strfreev(list);
	g_free(mask);

	return found;
}
