
/* : : generated by proto : : */

#if !defined(__PROTO__)
#include <prototyped.h>
#endif
#include "module.h"
#include "signals.h"
#include "misc.h"
#include "special-vars.h"
#include "settings.h"

#include "fe-windows.h"
#include "window-items.h"

#include "command-history.h"

/* command history */
static HISTORY_REC *global_history;
static int window_history;
static GSList *histories;

void command_history_add __PARAM__((HISTORY_REC *history, const char *text), (history, text)) __OTORP__(HISTORY_REC *history; const char *text;){
	GList *link;

	g_return_if_fail(history != NULL);
	g_return_if_fail(text != NULL);

	link = g_list_last(history->list);
	if (link != NULL && strcmp(link->data, text) == 0)
	  return; /* same as previous entry */

	if (settings_get_int("max_command_history") < 1 || 
	    history->lines < settings_get_int("max_command_history"))
		history->lines++;
	else {
		link = history->list;
		g_free(link->data);
		history->list = g_list_remove_link(history->list, link);
		g_list_free_1(link);
	}

	history->list = g_list_append(history->list, g_strdup(text));
}

HISTORY_REC *command_history_find __PARAM__((HISTORY_REC *history), (history)) __OTORP__(HISTORY_REC *history;){
	GSList *tmp;
	tmp = g_slist_find(histories, history);

	if (tmp == NULL)
		return NULL;
	else
		return tmp->data;
}

HISTORY_REC *command_history_find_name __PARAM__((const char *name), (name)) __OTORP__(const char *name;){
	GSList *tmp;

	if (name == NULL)
		return NULL;

	for (tmp = histories; tmp != NULL; tmp = tmp->next) {
		HISTORY_REC *rec = tmp->data;
		
		if (rec->name != NULL && g_strcasecmp(rec->name, name) == 0)
			return rec;
	}
	
	return NULL;
}

HISTORY_REC *command_history_current __PARAM__((WINDOW_REC *window), (window)) __OTORP__(WINDOW_REC *window;){
	HISTORY_REC *rec;

	if (window == NULL)
		return global_history;

	if (window_history)
		return window->history;

	rec = command_history_find_name(window->history_name);
	if (rec != NULL)
		return rec;

	return global_history;
}

const char *command_history_prev __PARAM__((WINDOW_REC *window, const char *text), (window, text)) __OTORP__(WINDOW_REC *window; const char *text;){
	HISTORY_REC *history;
	GList *pos;

	history = command_history_current(window);
	pos = history->pos;

	if (pos != NULL) {
		history->pos = history->pos->prev;
		if (history->pos == NULL)
                        history->over_counter++;
	} else {
		history->pos = g_list_last(history->list);
	}

	if (*text != '\0' &&
	    (pos == NULL || strcmp(pos->data, text) != 0)) {
		/* save the old entry to history */
		command_history_add(history, text);
	}

	return history->pos == NULL ? "" : history->pos->data;
}

const char *command_history_next __PARAM__((WINDOW_REC *window, const char *text), (window, text)) __OTORP__(WINDOW_REC *window; const char *text;){
	HISTORY_REC *history;
	GList *pos;

	history = command_history_current(window);
	pos = history->pos; 

	if (pos != NULL)
		history->pos = history->pos->next;
	else if (history->over_counter > 0) {
		history->over_counter--;
		history->pos = history->list;
	}

	if (*text != '\0' &&
	    (pos == NULL || strcmp(pos->data, text) != 0)) {
		/* save the old entry to history */
		command_history_add(history, text);
	}
	return history->pos == NULL ? "" : history->pos->data;
}

void command_history_clear_pos_func __PARAM__((HISTORY_REC *history, gpointer user_data), (history, user_data)) __OTORP__(HISTORY_REC *history; gpointer user_data;){
	history->over_counter = 0;
	history->pos = NULL;
}

void command_history_clear_pos __PARAM__((WINDOW_REC *window), (window)) __OTORP__(WINDOW_REC *window;){
	g_slist_foreach(histories, 
		       (GFunc) command_history_clear_pos_func, NULL);
}

HISTORY_REC *command_history_create __PARAM__((const char *name), (name)) __OTORP__(const char *name;){
	HISTORY_REC *rec;
	
	rec = g_new0(HISTORY_REC, 1);
	
	if (name != NULL)
		rec->name = g_strdup(name);

	histories = g_slist_append(histories, rec);
	
	return rec;
}

void command_history_destroy __PARAM__((HISTORY_REC *history), (history)) __OTORP__(HISTORY_REC *history;){
	g_return_if_fail(history != NULL);

	/* history->refcount should be 0 here, or somthing is wrong... */
	g_return_if_fail(history->refcount == 0);

	histories = g_slist_remove(histories, history);

	g_list_foreach(history->list, (GFunc) g_free, NULL);
	g_list_free(history->list);

	g_free_not_null(history->name);
	g_free(history);
}

void command_history_link __PARAM__((const char *name), (name)) __OTORP__(const char *name;){
	HISTORY_REC *rec;
	rec = command_history_find_name(name);

	if (rec == NULL)
		rec = command_history_create(name);

	rec->refcount++;
}

void command_history_unlink __PARAM__((const char *name), (name)) __OTORP__(const char *name;){
	HISTORY_REC *rec;
	rec = command_history_find_name(name);

	if (rec == NULL)
		return;

	if (--(rec->refcount) <= 0)
		command_history_destroy(rec);
}

static void sig_window_created __PARAM__((WINDOW_REC *window, int automatic), (window, automatic)) __OTORP__(WINDOW_REC *window; int automatic;){
	window->history = command_history_create(NULL);
}

static void sig_window_destroyed __PARAM__((WINDOW_REC *window), (window)) __OTORP__(WINDOW_REC *window;){
	command_history_unlink(window->history_name);
	command_history_destroy(window->history);
	g_free_not_null(window->history_name);
}

static void sig_window_history_changed __PARAM__((WINDOW_REC *window, const char *oldname), (window, oldname)) __OTORP__(WINDOW_REC *window; const char *oldname;){
	command_history_link(window->history_name);
	command_history_unlink(oldname);
}

static char *special_history_func __PARAM__((const char *text, __V_ *item, int *free_ret), (text, item, free_ret)) __OTORP__(const char *text; __V_ *item; int *free_ret;){
	WINDOW_REC *window;
	HISTORY_REC *history;
	GList *tmp;
        char *findtext, *ret;

	window = item == NULL ? active_win : window_item_window(item);

	findtext = g_strdup_printf("*%s*", text);
	ret = NULL;

	history = command_history_current(window);
	for (tmp = history->list; tmp != NULL; tmp = tmp->next) {
		const char *line = tmp->data;

		if (match_wildcards(findtext, line)) {
			*free_ret = TRUE;
                        ret = g_strdup(line);
		}
	}
	g_free(findtext);

	return ret;
}

static void read_settings __PARAM__((void), ()){
	window_history = settings_get_bool("window_history");
}

void command_history_init __PARAM__((void), ()){
	settings_add_int("history", "max_command_history", 100);
	settings_add_bool("history", "window_history", FALSE);

	special_history_func_set(special_history_func);

	global_history = command_history_create(NULL);

	read_settings();
	signal_add("window created", (SIGNAL_FUNC) sig_window_created);
	signal_add("window destroyed", (SIGNAL_FUNC) sig_window_destroyed);
	signal_add("window history changed", (SIGNAL_FUNC) sig_window_history_changed);
	signal_add("setup changed", (SIGNAL_FUNC) read_settings);
}

void command_history_deinit __PARAM__((void), ()){
	signal_remove("window created", (SIGNAL_FUNC) sig_window_created);
	signal_remove("window destroyed", (SIGNAL_FUNC) sig_window_destroyed);
	signal_remove("window history changed", (SIGNAL_FUNC) sig_window_history_changed);
	signal_remove("setup changed", (SIGNAL_FUNC) read_settings);

	command_history_destroy(global_history);
}
