
/* : : generated by proto : : */

#if !defined(__PROTO__)
#include <prototyped.h>
#endif
#include "module.h"
#include "signals.h"
#include "misc.h"
#include "settings.h"

#include "irc-servers.h"
#include "servers-redirect.h"

static int timeout_tag;

static void lag_get __PARAM__((IRC_SERVER_REC *server), (server)) __OTORP__(IRC_SERVER_REC *server;){
	g_get_current_time(&server->lag_sent);
	server->lag_last_check = time(NULL);

	server_redirect_event(server, "ping", 1, NULL, FALSE,
			      "lag ping error",
                              "event pong", "lag pong", NULL);
	irc_send_cmdv(server, "PING %s", server->real_address);
}

/* we didn't receive PONG for some reason .. try again */
static void lag_ping_error __PARAM__((IRC_SERVER_REC *server), (server)) __OTORP__(IRC_SERVER_REC *server;){
	lag_get(server);
}

static void lag_event_pong __PARAM__((IRC_SERVER_REC *server, const char *data,
			   const char *nick, const char *addr), (server, data, nick, addr)) __OTORP__(IRC_SERVER_REC *server; const char *data;
			   const char *nick; const char *addr;){
	GTimeVal now;

	g_return_if_fail(data != NULL);

	if (server->lag_sent.tv_sec == 0) {
		/* not expecting lag reply.. */
		return;
	}

	g_get_current_time(&now);
	server->lag = (int) get_timeval_diff(&now, &server->lag_sent);
	memset(&server->lag_sent, 0, sizeof(server->lag_sent));

	signal_emit("server lag", 1, server);
}

static void sig_unknown_command __PARAM__((IRC_SERVER_REC *server, const char *data), (server, data)) __OTORP__(IRC_SERVER_REC *server; const char *data;){
	char *params, *cmd;

	g_return_if_fail(data != NULL);

	params = event_get_params(data, 2, NULL, &cmd);
	if (g_strcasecmp(cmd, "PING") == 0) {
		/* some servers have disabled PING command, don't bother
		   trying alternative methods to detect lag with these
		   servers. */
		server->disable_lag = TRUE;
		server->lag_sent.tv_sec = 0;
                server->lag = 0;
	}
	g_free(params);
}

static int sig_check_lag __PARAM__((void), ()){
	GSList *tmp, *next;
	time_t now;
	int lag_check_time, max_lag;

	lag_check_time = settings_get_int("lag_check_time");
	max_lag = settings_get_int("lag_max_before_disconnect");

	if (lag_check_time <= 0)
		return 1;

	now = time(NULL);
	for (tmp = servers; tmp != NULL; tmp = next) {
		IRC_SERVER_REC *rec = tmp->data;

		next = tmp->next;
		if (!IS_IRC_SERVER(rec) || rec->disable_lag)
			continue;

		if (rec->lag_sent.tv_sec != 0) {
			/* waiting for lag reply */
			if (max_lag > 1 && now-rec->lag_sent.tv_sec > max_lag) {
				/* too much lag, disconnect */
				signal_emit("server lag disconnect", 1, rec);
				rec->connection_lost = TRUE;
				server_disconnect((SERVER_REC *) rec);
			}
		} else if (rec->lag_last_check+lag_check_time < now &&
			 rec->cmdcount == 0 && rec->connected) {
			/* no commands in buffer - get the lag */
			lag_get(rec);
		}
	}

	return 1;
}

void lag_init __PARAM__((void), ()){
	settings_add_int("misc", "lag_check_time", 60);
	settings_add_int("misc", "lag_max_before_disconnect", 300);

	timeout_tag = g_timeout_add(1000, (GSourceFunc) sig_check_lag, NULL);
	signal_add_first("lag pong", (SIGNAL_FUNC) lag_event_pong);
        signal_add("lag ping error", (SIGNAL_FUNC) lag_ping_error);
        signal_add("event 421", (SIGNAL_FUNC) sig_unknown_command);
}

void lag_deinit __PARAM__((void), ()){
	g_source_remove(timeout_tag);
	signal_remove("lag pong", (SIGNAL_FUNC) lag_event_pong);
        signal_remove("lag ping error", (SIGNAL_FUNC) lag_ping_error);
        signal_remove("event 421", (SIGNAL_FUNC) sig_unknown_command);
}
